#include <avr/io.h>
#include <stdint.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "USI_TWI_SLAVE.h"

#define TWI_Buff_size 15
unsigned char TWI_slaveAddress = 0x10;
unsigned char TWI_Buff[TWI_Buff_size];
unsigned char TWI_Buff_Index=0;


struct
{
  unsigned char AddressMode:1;
  unsigned char MasterReadDataMode:1;
  unsigned char PrepareMasterWriteDataMode:1;
  unsigned char MasterWriteDataMode:1;                  // masterReadDataMode = 0 is equal to "writeDataMode" = 1
  unsigned char GenAddressCall:1;
  unsigned char SlaveCheckAck:1;
  unsigned char SlaveSendAck:1;
  unsigned char unused:1;
} USI_TWI_State;

void USI_TWI_Slave_Initialise()
{
	TWI_Buff_Index=0;
	USI_TWI_State.AddressMode=FALSE;
	USI_TWI_State.MasterReadDataMode=FALSE;
	USI_TWI_State.MasterWriteDataMode=FALSE;
	USI_TWI_State.GenAddressCall=FALSE;
	USI_TWI_State.SlaveCheckAck=FALSE;

	PORT_USI |=  (1<<PORT_USI_SCL);                                 // Set SCL high
	PORT_USI |=  (1<<PORT_USI_SDA);                                 // Set SDA high
	DDR_USI  |=  (1<<PORT_USI_SCL);                                 // Set SCL as output
	DDR_USI  &= ~(1<<PORT_USI_SDA);                                 // Set SDA as input
	USISR    = 0xF0;
	USICR    =  (1<<USISIE)|(0<<USIOIE)|                            // Enable Start Condition Interrupt. Disable Overflow Interrupt.
              (1<<USIWM1)|(0<<USIWM0)|                            // Set USI in Two-wire mode. No USI Counter overflow prior
                                                                  // to first Start Condition (potentail failure)
              (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|                // Shift Register Clock Source = External, positive edge
              (0<<USITC);
}

void USI_TWI_Slave_WriteBytes(unsigned char *Msg, unsigned char MsgSize)
{
	unsigned char i=0;
	for(i=0;i<MsgSize;i++)
	{
		TWI_Buff[i]=Msg[i];
	}
	USIDR=TWI_Buff[0];
}

void USI_TWI_Slave_ReadBytes(unsigned char *Msg, unsigned char MsgSize)
{
	unsigned char i=0;
	for(i=0;i<MsgSize;i++)
	{
		Msg[i]=TWI_Buff[i];
	}
}

ISR(USI_START_vect) 
{   
	USI_TWI_State.AddressMode=TRUE;
	unsigned char tmpUSISR;                                         // Temporary variable to store volatile
    tmpUSISR = USISR;
    DDR_USI  &= ~(1<<PORT_USI_SDA);
	while ( (PIN_USI & (1<<PORT_USI_SCL)) & !(tmpUSISR & (1<<USIPF)));
	
	/* Enable TWI Coutner overflow and interrupts*/
    USICR   =   (1<<USISIE)|(1<<USIOIE)|                            // Enable Overflow and Start Condition Interrupt. (Keep StartCondInt to detect RESTART)
                (1<<USIWM1)|(0<<USIWM0)|                            // Set USI in Two-wire mode.
                (1<<USICS1)|(0<<USICS0)|(0<<USICLK)|                // Shift Register Clock Source = External, positive edge
                (0<<USITC);
    USISR   =   (1<<USISIF)|(1<<USIOIF)|(1<<USIPF)|(1<<USIDC)|      // Clear flags
                (0x0<<USICNT0);				                             // Set USI to sample 8 bits i.e. count 16 external pin toggles.	
}

ISR(USI_OVF_vect)
{

	if(USI_TWI_State.AddressMode)//Wenn nich keine Adresse erhalten wurde
	{
		if ( ((USIDR>>1) == TWI_slaveAddress)||(USIDR==0) ) //Gltigkeit der Adresse prfen
		{
			RecievedAddress=USIDR;
			if (USIDR == 0) USI_TWI_State.GenAddressCall = TRUE;
            else USI_TWI_State.GenAddressCall = FALSE;
			
			if ( USIDR & 0x01 ) USI_TWI_State.MasterReadDataMode=TRUE;
			else USI_TWI_State.PrepareMasterWriteDataMode=TRUE;

		  	SET_USI_TO_SEND_ACK();
			
			while ( !(USISR & ( (1<<USISIF)|(1<<USIOIF) )) );
			USISR |= (1<<USIOIF); //Das Aufgetretene Flag lschen
			DDR_USI &=  ~(1<<PORT_USI_SDA);
			USI_TWI_State.AddressMode=FALSE;
		}
		else
		{
			USI_TWI_Slave_Initialise();
            return;
		}
	}
		
	if(USI_TWI_State.SlaveCheckAck)
	{	
		SET_USI_TO_READ_ACK();
		while ( !(USISR & ( (1<<USISIF)|(1<<USIOIF) )) );
		USISR |= (1<<USIOIF); //Das Aufgetretene Flag lschen
			
		if (USIDR&0x01) // If NACK, the master does not want more data.
     	{
       		USI_TWI_Slave_Initialise();
       		return;
		}
		else
		{
		
			USI_TWI_State.SlaveCheckAck=FALSE;	
			USI_TWI_State.MasterReadDataMode=TRUE;
		}
	}
	
	if(USI_TWI_State.SlaveSendAck)
	{
		while ( !(USISR & ( (1<<USISIF)|(1<<USIOIF) )) );
		USISR |= (1<<USIOIF); //Das Aufgetretene Flag lschen
		DDR_USI &=  ~(1<<PORT_USI_SDA);
	}

	if(USI_TWI_State.MasterWriteDataMode)
	{
		TWI_Buff[TWI_Buff_Index]=USIDR;
		TWI_Buff_Index++;
		if(TWI_Buff_Index>TWI_Buff_size)
		{
			SET_USI_TO_SEND_NACK();
		}
		else 
		{
			SET_USI_TO_SEND_ACK();
		}
		USI_TWI_State.SlaveSendAck=TRUE;
		USI_TWI_State.MasterWriteDataMode=FALSE;
	}
	
	if(USI_TWI_State.MasterReadDataMode)
	{
		USIDR= TWI_Buff[TWI_Buff_Index];
		TWI_Buff_Index++;
		SET_USI_TO_SEND_DATA();
		USI_TWI_State.MasterReadDataMode=FALSE;
		USI_TWI_State.SlaveCheckAck=TRUE;		
	}
	if(USI_TWI_State.PrepareMasterWriteDataMode)
	{
		SET_USI_TO_READ_DATA();
		USI_TWI_State.MasterWriteDataMode=TRUE;
		USI_TWI_State.PrepareMasterWriteDataMode=FALSE;
	}

}
